// Generated from the OSSA cipher list
// version: 3.4
// date: 2024-04-11

#define TLS_CIPHERS_MANDATORY(X) \
  X("TLS_ECDHE_ECDSA_WITH_AES_128_GCM_SHA256", "ECDHE-ECDSA-AES128-GCM-SHA256") \
  X("TLS_ECDHE_ECDSA_WITH_AES_256_GCM_SHA384", "ECDHE-ECDSA-AES256-GCM-SHA384") \
  X("TLS_ECDHE_RSA_WITH_AES_128_GCM_SHA256", "ECDHE-RSA-AES128-GCM-SHA256") \

#define TLS_CIPHERS_APPROVED(X) \
  X("TLS_AES_128_GCM_SHA256", "") \
  X("TLS_AES_256_GCM_SHA384", "") \
  X("TLS_CHACHA20_POLY1305_SHA256", "") \
  X("TLS_AES_128_CCM_SHA256", "") \
  X("TLS_ECDHE_RSA_WITH_AES_256_GCM_SHA384", "ECDHE-RSA-AES256-GCM-SHA384") \
  X("TLS_ECDHE_ECDSA_WITH_CHACHA20_POLY1305_SHA256", "ECDHE-ECDSA-CHACHA20-POLY1305") \
  X("TLS_ECDHE_RSA_WITH_CHACHA20_POLY1305_SHA256", "ECDHE-RSA-CHACHA20-POLY1305") \
  X("TLS_ECDHE_ECDSA_WITH_AES_256_CCM", "ECDHE-ECDSA-AES256-CCM") \
  X("TLS_ECDHE_ECDSA_WITH_AES_128_CCM", "ECDHE-ECDSA-AES128-CCM") \

#define TLS_CIPHERS_UNACCEPTABLE(X) \
  X("TLS_ECDH_anon_WITH_NULL_SHA", "AECDH-NULL-SHA") \
  X("TLS_ECDHE_RSA_WITH_NULL_SHA", "ECDHE-RSA-NULL-SHA") \
  X("TLS_ECDHE_ECDSA_WITH_NULL_SHA", "ECDHE-ECDSA-NULL-SHA") \
  X("TLS_GOSTR341001_WITH_NULL_GOSTR3411", "GOST94-NULL-GOST94") \
  X("TLS_GOSTR341094_WITH_NULL_GOSTR3411", "GOST2001-GOST89-GOST89") \
  X("TLS_ECDH_RSA_WITH_NULL_SHA", "ECDH-RSA-NULL-SHA") \
  X("TLS_ECDH_ECDSA_WITH_NULL_SHA", "ECDH-ECDSA-NULL-SHA") \
  X("TLS_RSA_WITH_NULL_SHA256", "NULL-SHA256") \
  X("TLS_RSA_WITH_NULL_SHA", "NULL-SHA") \
  X("TLS_RSA_WITH_NULL_MD5", "NULL-MD5") \
  X("TLS_ECDH_anon_WITH_AES_256_CBC_SHA", "AECDH-AES256-SHA") \
  X("TLS_DH_anon_WITH_AES_256_GCM_SHA384", "ADH-AES256-GCM-SHA384") \
  X("TLS_DH_anon_WITH_AES_256_CBC_SHA256", "ADH-AES256-SHA256") \
  X("TLS_DH_anon_WITH_AES_256_CBC_SHA", "ADH-AES256-SHA") \
  X("TLS_DH_anon_WITH_CAMELLIA_256_CBC_SHA256", "ADH-CAMELLIA256-SHA256") \
  X("TLS_DH_anon_WITH_CAMELLIA_256_CBC_SHA", "ADH-CAMELLIA256-SHA") \
  X("TLS_ECDH_anon_WITH_AES_128_CBC_SHA", "AECDH-AES128-SHA") \
  X("TLS_DH_anon_WITH_AES_128_GCM_SHA256", "ADH-AES128-GCM-SHA256") \
  X("TLS_DH_anon_WITH_AES_128_CBC_SHA256", "ADH-AES128-SHA256") \
  X("TLS_DH_anon_WITH_AES_128_CBC_SHA", "ADH-AES128-SHA") \
  X("TLS_DH_anon_WITH_CAMELLIA_128_CBC_SHA256", "ADH-CAMELLIA128-SHA256") \
  X("TLS_DH_anon_WITH_CAMELLIA_128_CBC_SHA", "AADH-CAMELLIA128-SHA") \
  X("TLS_ECDH_anon_WITH_RC4_128_SHA", "AECDH-RC4-SHA") \
  X("TLS_DH_anon_WITH_RC4_128_MD5", "ADH-RC4-MD5") \
  X("TLS_ECDH_anon_WITH_3DES_EDE_CBC_SHA", "AECDH-DES-CBC3-SHA") \
  X("TLS_DH_anon_WITH_3DES_EDE_CBC_SHA", "ADH-DES-CBC3-SHA") \
  X("TLS_DH_anon_WITH_DES_CBC_SHA", "ADH-DES-CBC-SHA") \
  X("SSL_RSA_EXPORT_WITH_RC4_40_MD5", "EXP-RC4-MD5") \
  X("SSL_RSA_EXPORT_WITH_RC2_CBC_40_MD5", "EXP-RC2-CBC-MD5") \
  X("SSL_RSA_EXPORT_WITH_DES40_CBC_SHA", "EXP-DES-CBC-SHA") \
  X("SSL_DH_DSS_EXPORT_WITH_DES40_CBC_SHA", "") \
  X("SSL_DH_RSA_EXPORT_WITH_DES40_CBC_SHA", "") \
  X("TLS_DH_DSS_EXPORT_WITH_DES40_CBC_SHA", "EXP-DH-DSS-DES-CBC-SHA") \
  X("TLS_DH_RSA_EXPORT_WITH_DES40_CBC_SHA", "EXP-DH-RSA-DES-CBC-SHA") \
  X("SSL_DHE_DSS_EXPORT_WITH_DES40_CBC_SHA", "EXP-EDH-DSS-DES-CBC-SHA") \
  X("SSL_DHE_RSA_EXPORT_WITH_DES40_CBC_SHA", "EXP-EDH-RSA-DES-CBC-SHA") \
  X("SSL_DH_anon_EXPORT_WITH_RC4_40_MD5", "EXP-ADH-RC4-MD5") \
  X("SSL_DH_anon_EXPORT_WITH_DES40_CBC_SHA", "EXP-ADH-DES-CBC-SHA") \
  X("TLS_KRB5_EXPORT_WITH_DES_CBC_40_SHA", "EXP-KRB5-DES-CBC-SHA") \
  X("TLS_KRB5_EXPORT_WITH_RC2_CBC_40_SHA", "EXP-KRB5-RC2-CBC-SHA") \
  X("TLS_KRB5_EXPORT_WITH_RC4_40_SHA", "EXP-KRB5-RC4-SHA") \
  X("TLS_KRB5_EXPORT_WITH_DES_CBC_40_MD5", "EXP-KRB5-DES-CBC-MD5") \
  X("TLS_KRB5_EXPORT_WITH_RC2_CBC_40_MD5", "EXP-KRB5-RC2-CBC-MD5") \
  X("TLS_KRB5_EXPORT_WITH_RC4_40_MD5", "EXP-KRB5-RC4-MD5") \
  X("TLS_RSA_EXPORT_WITH_RC4_40_MD5", "EXP-RC4-MD5") \
  X("TLS_RSA_EXPORT_WITH_RC2_CBC_40_MD5", "EXP-RC2-CBC-MD5") \
  X("TLS_RSA_EXPORT_WITH_DES40_CBC_SHA", "") \
  X("TLS_DHE_DSS_EXPORT_WITH_DES40_CBC_SHA", "EXP-EDH-DSS-DES-CBC-SHA") \
  X("TLS_DHE_RSA_EXPORT_WITH_DES40_CBC_SHA", "EXP-EDH-RSA-DES-CBC-SHA") \
  X("TLS_DH_anon_EXPORT_WITH_RC4_40_MD5", "EXP-ADH-RC4-MD5") \
  X("TLS_DH_anon_EXPORT_WITH_DES40_CBC_SHA", "EXP-ADH-DES-CBC-SHA") \
  X("TLS_RSA_EXPORT1024_WITH_DES_CBC_SHA", "EXP1024-DES-CBC-SHA") \
  X("TLS_RSA_EXPORT1024_WITH_RC4_56_SHA", "EXP1024-RC4-SHA") \
  X("TLS_RSA_EXPORT1024_WITH_RC4_56_MD5", "EXP1024-RC4-MD5") \
  X("TLS_RSA_EXPORT1024_WITH_RC2_CBC_56_MD5", "EXP1024-RC2-CBC-MD5") \
  X("TLS_DHE_DSS_EXPORT1024_WITH_DES_CBC_SHA", "EXP1024-DHE-DSS-DES-CBC-SHA") \
  X("TLS_DHE_DSS_EXPORT1024_WITH_RC4_56_SHA", "EXP1024-DHE-DSS-RC4-SHA") \
  X("SSL_CK_RC4_128_EXPORT40_WITH_MD5", "EXP-RC4-MD5") \
  X("SSL_CK_RC2_128_CBC_EXPORT40_WITH_MD5", "EXP-RC2-CBC-MD5") \
  X("SSL_CK_RC2_128_CBC_EXPORT40_WITH_MD5", "EXP-RC2-MD5") \
  X("TLS_DHE_RSA_WITH_DES_CBC_SHA", "EDH-RSA-DES-CBC-SHA") \
  X("TLS_DHE_DSS_WITH_DES_CBC_SHA", "EDH-DSS-DES-CBC-SHA") \
  X("TLS_DH_anon_WITH_DES_CBC_SHA", "ADH-DES-CBC-SHA") \
  X("SL_CK_DES_64_CBC_WITH_SHA", "DES-CBC-SHA") \
  X("TLS_KRB5_WITH_DES_CBC_SHA", "KRB5-DES-CBC-SHA") \
  X("TLS_DH_anon_WITH_RC4_128_MD5", "ADH-RC4-MD5") \
  X("TLS_RSA_WITH_RC4_128_MD5", "RC4-MD5") \
  X("TLS_RSA_WITH_NULL_MD5", "NULL-MD5") \
  X("TLS_KRB5_WITH_DES_CBC_MD5", "KRB5-DES-CBC-MD5") \
  X("TLS_KRB5_WITH_RC4_128_MD5", "KRB5-RC4-MD5") \
  X("TLS_KRB5_WITH_IDEA_CBC_MD5", "KRB5-IDEA-CBC-MD5") \
  X("TLS_ECDHE_RSA_WITH_RC4_128_SHA", "ECDHE-RSA-RC4-SHA") \
  X("TLS_ECDHE_ECDSA_WITH_RC4_128_SHA", "ECDHE-ECDSA-RC4-SHA") \
  X("TLS_ECDH_anon_WITH_RC4_128_SHA", "AECDH-RC4-SHA") \
  X("TLS_ECDH_RSA_WITH_RC4_128_SHA", "ECDH-RSA-RC4-SHA") \
  X("TLS_ECDH_ECDSA_WITH_RC4_128_SHA", "ECDH-ECDSA-RC4-SHA") \
  X("TLS_RSA_WITH_RC4_128_SHA", "RC4-SHA") \
  X("TLS_KRB5_WITH_RC4_128_SHA", "KRB5-RC4-SHA") \
  X("TLS_ECDH_anon_WITH_NULL_SHA", "AECDH-NULL-SHA") \
  X("TLS_ECDH_RSA_WITH_NULL_SHA", "ECDH-RSA-NULL-SHA") \
  X("TLS_ECDH_ECDSA_WITH_NULL_SHA", "ECDH-ECDSA-NULL-SHA") \
  X("TLS_PSK_WITH_AES_256_CBC_SHA", "PSK-AES256-CBC-SHA") \
  X("TLS_PSK_WITH_AES_128_CBC_SHA", "PSK-AES128-CBC-SHA") \
  X("TLS_PSK_WITH_3DES_EDE_CBC_SHA", "PSK-3DES-EDE-CBC-SHA") \
  X("TLS_PSK_WITH_RC4_128_SHA", "PSK-RC4-SHA") \
  X("TLS_RSA_EXPORT_WITH_RC2_CBC_40_MD5", "EXP-RC2-CBC-MD5") \
  X("TLS_KRB5_EXPORT_WITH_RC2_CBC_40_SHA", "EXP-KRB5-RC2-CBC-SHA") \
  X("TLS_RSA_EXPORT1024_WITH_RC2_CBC_56_MD5", "EXP1024-RC2-CBC-MD5") \
  X("SSL_CK_RC2_128_CBC_WITH_MD5", "RC2-CBC-MD5") \
  X("SSL_CK_RC2_128_CBC_EXPORT40_WITH_MD5", "EXP-RC2-CBC-MD5") \
  X("TLS_DH_RSA_WITH_AES_128_CBC_SHA256", "DH-RSA-AES128-SHA256") \
  X("TLS_DH_RSA_WITH_AES_256_CBC_SHA256", "DH-RSA-AES256-SHA256") \
  X("TLS_DH_DSS_WITH_AES_128_CBC_SHA256", "DH-DSS-AES128-SHA256") \
  X("TLS_DH_DSS_WITH_AES_128_CBC_SHA", "DH-DSS-AES128-SHA") \
  X("TLS_DH_DSS_WITH_AES_256_CBC_SHA", "DH-DSS-AES256-SHA") \
  X("TLS_DH_DSS_WITH_AES_256_CBC_SHA256", "DH-DSS-AES256-SHA256") \
  X("TLS_DH_RSA_WITH_AES_128_CBC_SHA", "DH-RSA-AES128-SHA") \
  X("TLS_DH_RSA_WITH_AES_256_CBC_SHA", "DH-RSA-AES256-SHA") \
  X("TLS_DH_DSS_WITH_AES_128_GCM_SHA256", "DH-DSS-AES128-GCM-SHA256") \
  X("TLS_DH_DSS_WITH_AES_256_GCM_SHA384", "DH-DSS-AES256-GCM-SHA384") \
  X("TLS_DH_RSA_WITH_AES_128_GCM_SHA256", "DH-RSA-AES128-GCM-SHA256") \
  X("TLS_DH_RSA_WITH_AES_256_GCM_SHA384", "DH-RSA-AES256-GCM-SHA384") \
  X("TLS_DH_DSS_WITH_3DES_EDE_CBC_SHA", "DH-DSS-DES-CBC3-SHA") \
  X("TLS_DH_RSA_WITH_3DES_EDE_CBC_SHA", "DH-RSA-DES-CBC3-SHA") \
  X("TLS_DHE_DSS_WITH_3DES_EDE_CBC_SHA", "EDH-DSS-DES-CBC3-SHA") \
  X("TLS_DHE_RSA_WITH_3DES_EDE_CBC_SHA", "EDH-RSA-DES-CBC3-SHA") \
  X("TLS_ECDH_RSA_WITH_3DES_EDE_CBC_SHA", "ECDH-RSA-DES-CBC3-SHA") \
  X("TLS_ECDH_ECDSA_WITH_3DES_EDE_CBC_SHA", "ECDH-ECDSA-DES-CBC3-SHA") \
  X("TLS_ECDHE_RSA_WITH_3DES_EDE_CBC_SHA", "ECDHE-RSA-DES-CBC3-SHA") \
  X("TLS_ECDHE_ECDSA_WITH_3DES_EDE_CBC_SHA", "ECDHE-ECDSA-DES-CBC3-SHA") \
  X("TLS_RSA_WITH_3DES_EDE_CBC_SHA", "DES-CBC3-SHA") \
  X("TLS_KRB5_WITH_3DES_EDE_CBC_SHA", "KRB5-DES-CBC3-SHA") \
  X("TLS_KRB5_WITH_3DES_EDE_CBC_MD5", "KRB5-DES-CBC3-MD5") \
  X("TLS_KRB5_WITH_IDEA_CBC_SHA", "KRB5-IDEA-CBC-SHA") \

// This comment is here to avoid "backslash-newline at end of file" compile warning.
