/*
 * Copyright (c) 2015, 2024, Oracle and/or its affiliates.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License, version 2.0, as
 * published by the Free Software Foundation.
 *
 * This program is designed to work with certain software (including
 * but not limited to OpenSSL) that is licensed under separate terms, as
 * designated in a particular file or component or in included license
 * documentation. The authors of MySQL hereby grant you an additional
 * permission to link the program and your derivative works with the
 * separately licensed software that they have either included with
 * the program or referenced in the documentation.
 *
 * Without limiting anything contained in the foregoing, this file,
 * which is part of Connector/C++, is also subject to the
 * Universal FOSS Exception, version 1.0, a copy of which can be found at
 * https://oss.oracle.com/licenses/universal-foss-exception.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License, version 2.0, for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
  Unit tests for CDK time functions.
*/

#include "test.h"
#include <iostream>
#include <mysql/cdk/foundation/cdk_time.h>

using ::std::cout;
using ::std::endl;
using namespace cdk::foundation;


TEST(Foundation, time)
{
  using cdk::foundation::time_t;
  using cdk::foundation::sleep;

  time_t begin = get_time();

  cout <<"Start time is " <<begin/1000.0 <<" seconds from epoch"  <<endl;

  cout <<"Sleeping for 300 ms" <<endl;
  time_t start = get_time();
  sleep(300);
  time_t stop = get_time();

  cout <<"Measured delay: " <<(stop - start) <<"ms" <<endl;

  // Note: observed accuracy worse than +-30ms
  #define MARGIN 30
  EXPECT_FALSE(stop > start + 300 + MARGIN);
  EXPECT_FALSE(stop < start + 300 - MARGIN);
}

